#include <algorithm>
#include <limits>
#include <iostream>
#include <fstream>

#include "Sphere.h"
#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

// determine the closest hit point from a given ray on this object
// if no ray intersection exists, return maximum possible integer value
float Sphere::hit(ray& r) {
	glm::vec4 e = glm::vec4(r.getPos(), 1.0f);	//eye
	glm::vec4 d = glm::vec4(r.getDir(), 0.0f);	//direction
	if (hasTransforms) {
		//glm::mat4 mInv = glm::inverse(ctm);
		e = mInv * e;
		d = mInv * d;
	}

	//quadratic variable setup
	float A = glm::dot(d, d);//d.dot(d);
	float B = glm::dot(d, e - glm::vec4(pos, 1.0f));//d.dot(e - pos);
	glm::vec4 ecDiff = e - glm::vec4(pos, 1.0f);
	//float C = ecDiff.dot(ecDiff) - powf(radius, 2.0f);
	float C = glm::dot(ecDiff, ecDiff) - powf(radius, 2.0f);
	
	//default return value (no intersection)
	float defaultDist = numeric_limits<float>::max();

	//inner = value inside square root of quadratic eqn.
	//calculating this will determine no intersection/1 intersection/2 intersections
	float inner = powf(B, 2.0f) - (A * C);

	float result = defaultDist;

	if (inner < 0)	//no intersection
		return result;
	else if (inner == 0) {	//1 intersection
		//cout << "hit once" << endl;
		result = -B / A;
	}
	else {//2 intersections
		//cout << "hit twice" << endl;
		float t1 = (-B + sqrtf(inner)) / A;
		float t2 = (-B - sqrtf(inner)) / A;
		bool inside = (powf((e.x - pos.x), 2) + powf((e.y - pos.y), 2) + powf((e.z - pos.z), 2)) < (radius * radius);
		if (inside && this->getMat()->getType() == "dialectric")
			result = max(t1, t2);
		else
			result = min(t1, t2);	//return closest of the two intersections
	}

	if (result > 0)	//we only care about objects in front of the ray, not behind
		return result;
	return defaultDist;
}

float Sphere::hit(ray& r, bool inside) {
	glm::vec3 e = r.getPos();	//eye
	glm::vec3 d = r.getDir();	//direction

	//quadratic variable setup
	float A = glm::dot(d, d);//d.dot(d);
	float B = glm::dot(d, glm::vec3(e - pos));//d.dot(e - pos);
	glm::vec3 ecDiff = e - pos;
	//float C = ecDiff.dot(ecDiff) - powf(radius, 2.0f);
	float C = glm::dot(ecDiff, ecDiff) - powf(radius, 2.0f);
	//default return value (no intersection)
	float defaultDist = numeric_limits<float>::max();

	//inner = value inside square root of quadratic eqn.
	//calculating this will determine no intersection/1 intersection/2 intersections
	float inner = powf(B, 2.0f) - (A * C);

	float result = defaultDist;

	if (inner < 0)	//no intersection
		return result;
	else if (inner == 0) {	//1 intersection
		//cout << "hit once" << endl;
		result = -B / A;
	}
	else {//2 intersections
		//cout << "hit twice" << endl;
		float t1 = (-B + sqrtf(inner)) / A;
		float t2 = (-B - sqrtf(inner)) / A;
		if (inside)
			result = max(t1, t2);	//return closest of the two intersections
		else
			result = min(t1, t2);
	}

	if (result > 0)	//we only care about objects in front of the ray, not behind
		return result;
	return defaultDist;
}

void Sphere::createTexture(string inFile, img& tex) {
	// Load texture
	int w, h, ncomps;
	stbi_set_flip_vertically_on_load(true);
	tex.data = stbi_load(inFile.c_str(), &w, &h, &ncomps, 0);
	tex.numComp = ncomps;
	if (!tex.data) {
		cerr << inFile << " not found" << endl;
	}
	if (ncomps != 3) {
		cerr << inFile << " must have 3 components (RGB)" << endl;
	}
	if ((w & (w - 1)) != 0 || (h & (h - 1)) != 0) {
		cerr << inFile << " must be a power of 2" << endl;
	}
	tex.width = w;
	tex.height = h;
	tex.filename = inFile;
}