#pragma once

#ifndef SPHERE_H
#define SPHERE_H

#include "Shape.h"

using namespace std;

#define PI 3.1415f

struct img {
	unsigned char* data;
	string filename;
	int width;
	int height;
	int numComp;
} typedef img;

class Sphere : public Shape
{
public:
	Sphere() : Shape{} {										//
		pos = glm::vec3(0);										//
		baseColor = glm::vec3(0.5f);							//
		radius = 1.0f;											//
		mat = make_shared<Lambert>();							//
		hasTransforms = false;									//
		hasTexture = false;										//
	}															//
	Sphere(glm::vec3 p, glm::vec3 rgb, float r) : Shape{} {		//	unnecessary at this point, all spheres will have some material
		pos = p;												//
		baseColor = rgb;	//scale from decimal to 0-255		//
		radius = r;												//
		mat = make_shared<Lambert>();							//
		hasTransforms = false;									//
		hasTexture = false;										//
	}															//
	Sphere(glm::vec3 p, glm::vec3 rgb, float r, shared_ptr<Material> m) : Shape{} {													//no transforms, no texture, no bumpMap
		pos = p;
		baseColor = rgb;	//scale from decimal to 0-255
		radius = r;
		mat = m;
		hasTransforms = false;
		mInv = glm::mat4(1.0f);
		hasTexture = false;
		hasBumps = false;
	}
	Sphere(glm::vec3 p, glm::vec3 rgb, float r, shared_ptr<Material> m, glm::mat4 transforms) : Shape{} {							//transforms, no texture, no bumpMap
		pos = p;
		baseColor = rgb;	//scale from decimal to 0-255
		radius = r;
		mat = m;
		hasTransforms = true;
		ctm = transforms;
		mInv = glm::inverse(ctm);
		hasTexture = false;
		hasBumps = false;
	}
	Sphere(glm::vec3 p, glm::vec3 rgb, float r, shared_ptr<Material> m, vector<string> texPath, bool hasTex, bool hasB) : Shape{} {	//no transforms, texture and/or bumpMap
		pos = p;
		baseColor = rgb;	//scale from decimal to 0-255
		radius = r;
		mat = m;
		hasTransforms = false;
		mInv = glm::mat4(1.0f);
		setImgs(texPath, hasTex, hasB);

		/*hasTexture = true;
		createTexture(texPath, texture);
		hasBumps = false;*/
	}
	Sphere(glm::vec3 p, glm::vec3 rgb, float r, shared_ptr<Material> m, glm::mat4 transforms, vector<string> texPath, bool hasTex, bool hasB) : Shape{} {			//transforms, texture and/or bumpMap
		pos = p;
		baseColor = rgb;	//scale from decimal to 0-255
		radius = r;
		mat = m;
		hasTransforms = true;
		ctm = transforms;
		mInv = glm::inverse(ctm);

		setImgs(texPath, hasTex, hasB);

		/*hasTexture = true;
		createTexture(texPath, texture);
		hasBumps = false;*/
	}

	void setImgs(vector<string> imgPaths, bool hasTex, bool hasB) {
		if (hasTex && !hasB) {		//tex, no bumpmap
			hasTexture = true;
			hasBumps = false;
			createTexture(imgPaths.at(0), texture);
		}
		else if (!hasTex && hasB) {	//no tex, bumpmap
			hasTexture = false;
			hasBumps = true;
			createTexture(imgPaths.at(0), bumpMap);
		}
		else if (hasTex && hasB) {	//tex, bumpmap
			hasTexture = true;
			hasBumps = true;
			createTexture(imgPaths.at(0), texture);
			createTexture(imgPaths.at(1), bumpMap);
		}
	}

	bool hasTex() {
		return hasTexture;
	}
	virtual float hit(ray& r);
	float hit(ray& r, bool inside);
	virtual string getType() {
		return type;
	}
	virtual float getRadius() {
		return radius;
	}
	virtual glm::vec3 getNorm(glm::vec3& pt) {
		glm::vec3 n = glm::normalize(glm::vec3(pt - pos));

		if (hasBumps && !hasTransforms) {
			//calculate new normal based on map
			//n = imgLookUp(bumpMap, n);

			n = calcBump(bumpMap, n);
		}

		if (hasTransforms) {
			//multiply contact pt by inverse transpose ctm
			glm::vec3 tempPt = glm::vec3(mInv * glm::vec4(pt, 1.0f));
			n = glm::normalize(glm::vec3(tempPt - pos));

			if (hasBumps) {
				//calculate new normal based on map
				//n = imgLookUp(bumpMap, n);

				n = calcBump(bumpMap, n);
			}

			glm::mat4 mInvT = glm::transpose(mInv);
			n = glm::vec3(mInvT * glm::vec4(n, 1.0f));
		}

		return n;
	}
	glm::vec3 getNormSp(ray& r) {
		float tVal = hit(r);
		glm::vec3 pt = r.getPos() + tVal * r.getDir();
		return getNorm(pt);
	}
	void createTexture(string inFile, img& tex);
	glm::vec3 getRGB(glm::vec3& pt) {
		if (!hasTexture) {
			return baseColor;
		}
		glm::vec3 n = getNorm(pt);

		glm::vec3 tempPt = glm::vec3(mInv * glm::vec4(pt, 1.0f));
		n = glm::normalize(glm::vec3(tempPt - pos));


		glm::vec3 retColor = imgLookUp(texture, n);
		return retColor;
	}

	glm::vec3 imgLookUp(img tex, glm::vec3 n) {
		float u = asin(n.x) / PI + 0.5f;
		float v = asin(n.y) / PI + 0.5f;
		int tu = ((u)*tex.width);
		int tv = ((v)*tex.height);
		int idx = floor(tex.numComp * (tv * tex.width + tu));
		return glm::vec3(tex.data[idx] / 255.0f, tex.data[idx + 1] / 255.0f, tex.data[idx + 2] / 255.0f);
	}

	glm::vec3 calcBump(img tex, glm::vec3 n) {
		glm::vec3 t = glm::cross(n, glm::vec3(0, 1, 0));
		if (!glm::length(t))
			t = glm::cross(n, glm::vec3(0, 0, 1));
		t = glm::normalize(t);
		glm::vec3 b = glm::normalize(glm::cross(n, t));
		glm::vec3 mapN = glm::normalize(imgLookUp(tex, n));
		mapN = (mapN * 2.0f) - glm::vec3(1.0f);
		glm::mat3 tbn = glm::mat3(t, b, n);
		return glm::normalize(tbn * mapN);
	}
private:
	string type = "sphere";
	float radius;
	glm::mat4 ctm;
	glm::mat4 mInv;
	bool hasTransforms;

	bool hasTexture;
	img texture;

	bool hasBumps;
	img bumpMap;
};

#endif // !SPHERE_H