/*
*	Matthew Key
*	CSC 473 Fall 2020
*	Final Project - Edge detection, normal mapping, and depth of field
*	Takes image dimensions, input pov file, and output ppm file as parameters
*	Produces output ppm file of basic ray-trace image
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <limits>
#include <memory>
#include <chrono>
#include <random>
#include <ctime>

#include "tiny_obj_loader.h"
#include "glm/glm.hpp"
#include "glm/gtx/string_cast.hpp"
#include "glm/vec3.hpp"
#include "utils.h"
#include "Shape.h"
#include "Sphere.h"
#include "Plane.h"
#include "Rectangle.h"
#include "Obj.h"
#include "implicit.h"
#include "asciiR.h"
#include "ppm.h"
#include "ray.h"
#include "camera.h"
#include "Parser.h"
#include "Material.h"
#include "Lambert.h"
#include "Specular.h"
#include "Dialectric.h"

using namespace std;

// takes in a file of ints and stores them in a vector
// stops reading at eof or at a non-int char
// returns the vector of ints
vector<int> readInts(istream& is) {
	vector<int> tmp;
	int inI;
	while (is >> inI)
		tmp.push_back(inI);
	return tmp;
}

// generate a ray given the appropriate parameters based on the camera position
ray genRay(glm::vec3 camPos, glm::vec3 LLC, glm::vec3 horiz, glm::vec3 vert, int width, int height, float x, float y) {
	float hp = x / (float)width;
	float vp = y / (float)height;

	/*hp = 0.5f;
	vp = 0.5f;*/

	glm::vec3 Dir = LLC + (horiz * hp) + (vert * vp) - camPos;
	Dir = glm::normalize(Dir);
	return ray(camPos, Dir);
}

glm::vec3 phong(camera& cam, Light& l, ray& r, vector<shared_ptr<Shape>>& objVec) {
	glm::vec3 returnColor = glm::vec3(0.0f);

	//find which object has been hit closest
	float closestPt = numeric_limits<float>::max();
	float hitDistance = closestPt;
	bool hitSomething = false;
	shared_ptr<Shape> hitShape;
	for (auto& s : objVec) {
		hitDistance = numeric_limits<float>::max();
		hitDistance = s->hit(r);
		if (hitDistance < closestPt) {
			closestPt = hitDistance;
			hitSomething = true;
			hitShape = s;
		}
	}

	if (hitSomething) {
		//calculate phong

		//find contact point
		glm::vec3 contactPt = r.getPos() + r.getDir() * closestPt;
		//find normal of surface at point
		string objType = hitShape->getType();
		glm::vec3 norm = glm::vec3(0, 1, 0);
		norm = hitShape->getNorm(contactPt);
		//norm.normalize();
		norm = glm::normalize(norm);
		contactPt += norm * 0.0001f;

		//ambient
		float ambientStrength = 0.1;
		glm::vec3 ambient = l.getColor() * ambientStrength;

		glm::vec3 lightDir = l.getPos() - contactPt;
		float distanceToLight = glm::length(lightDir);
		lightDir = glm::normalize(lightDir);
		ray toLight = ray(contactPt, lightDir);
		hitSomething = false;
		closestPt = distanceToLight;
		hitDistance = closestPt;
		for (auto& s : objVec) {
			hitDistance = distanceToLight;
			hitDistance = s->hit(toLight);
			if (hitDistance < distanceToLight) {
				closestPt = hitDistance;
				hitSomething = true;
				break;
			}
		}

		if (hitSomething) {	//shadow
			return hitShape->getRGB(contactPt) * ambient;
		}
		else {	//diffuse
			glm::vec3 diffuse = l.getColor() * max(glm::dot(norm, lightDir), 0.0f);
			glm::vec3 c = hitShape->getRGB(contactPt);
			return c * (ambient + diffuse);
		}
	}
	else {
		glm::vec3 unitDir = r.getDir();
		//unitDir.normalize();
		unitDir = glm::normalize(unitDir);
		float t = 0.5 * (unitDir.y + 1.0);
		return glm::vec3(1.0f) * (1.0f - t) + glm::vec3(0.0f, 0.27f, 0.67f) * t;
	}

	return returnColor;
}

bool edgeDetect(ray r, shared_ptr<Shape> obj, float d, vector<shared_ptr<Shape>>& objVec, glm::vec3 n) {
	float x, y, angle, hitDist, altHitDist, rise, run;
	ray newRay;
	glm::vec3 newPos;
	glm::vec3 contactPt = r.getPos() + r.getDir() * d;
	glm::vec3 altPt;
	float slopes[8];
	float dists[8];
	glm::vec3 norms[8];
	glm::vec3 hitNorm;
	for (int i = 0; i < 8; i++) {
		angle = 45.0f * i;
		x = cos(angle) * 0.05f;
		y = sin(angle) * 0.05f;
		newPos = r.getPos();
		newPos.x += x;
		newPos.y += y;
		newRay = ray(newPos, r.getDir());
		hitDist = obj->hit(newRay);
		hitNorm = obj->getNorm(newPos);	//using newPos as dummy point shouldn't matter since hit sets the previously hit normal
		if (!(hitDist < numeric_limits<float>::max()))
			return true;

		if (abs(hitDist - d) > (0.05f * d))
			return true;

		for (auto& s : objVec) {
			if (s != obj) {
				altHitDist = s->hit(r);
				if (altHitDist < hitDist) {
					return true;
				}
			}
		}

		norms[i] = hitNorm;
		dists[i] = hitDist;

		// attempt at crease edge detection

		/*altPt = newRay.getPos() + newRay.getDir() * hitDist;
		run = sqrt(pow(contactPt.x - altPt.x, 2.0f) + pow(contactPt.y - altPt.y, 2.0f));
		rise = altPt.z - contactPt.z;
		slopes[i] = rise / run;*/
	}

	// below are attempts at crease edge detection
	/*float mean = 0.0f;
	for (int i = 0; i < 8; i++) {
		mean += dists[i];
	}
	mean /= 8.0f;
	float stdDev = 0.0f;
	for (int i = 0; i < 8; i++) {
		stdDev += pow((dists[i] - mean), 2.0f);
	}
	stdDev /= 8.0f;
	stdDev = sqrt(stdDev);
	for (int i = 0; i < 8; i++) {
		if (dists[i] < (mean - 2.0f * stdDev) || dists[i] > (mean + 2.0f * stdDev))
			return true;
	}*/

	/*for (int i = 0; i < 8; i++) {
		for (int j = i + 1; j < 8; j++) {
			if (abs(slopes[i] - slopes[j]) > 50.0f)
				return true;
		}
	}*/
	return false;
}

glm::vec3 foggy(camera& cam, Light& l, ray& r, vector<shared_ptr<Shape>>& objVec, int depth, bool hasEm, bool firstBounce) {
	float closestPt = numeric_limits<float>::max();
	bool hitSomething = false;
	float hitDist = closestPt;
	if (depth <= 0)
		return glm::vec3(0.0f);
	glm::vec3 retColor(0.0f);
	shared_ptr<Shape> hitShape;
	for (auto& s : objVec) {
		hitDist = s->hit(r);
		if (hitDist < closestPt) {
			closestPt = hitDist;
			hitSomething = true;
			hitShape = s;
		}
	}
	if (hitSomething) {
		glm::vec3 contactPt = r.getPos() + r.getDir() * closestPt;
		if (hitShape->getMat()->getType() != "emissive") {
			glm::vec3 n = hitShape->getNorm(contactPt);
			
			if(firstBounce)
				if (edgeDetect(r, hitShape, closestPt, objVec, n))
					return glm::vec3(0);

			ray scatRay = hitShape->getMat()->scatter(r, n, contactPt);
			glm::vec3 c = hitShape->getRGB(contactPt);
			retColor = c * foggy(cam, l, scatRay, objVec, depth - 1, hasEm, false);
			
			// used for rendering normals
			//retColor = hitShape->getNorm(contactPt) * foggy(cam, l, scatRay, objVec, depth - 1, hasEm);
			//retColor = n;
			return retColor;
		}
		return hitShape->getRGB(contactPt);
	}
	else {
		if (!hasEm) {
			glm::vec3 unitDir = r.getDir();
			unitDir = glm::normalize(unitDir);
			float t = 0.5 * (unitDir.y + 1.0);
			return glm::vec3(1.0f) * (1.0f - t) + glm::vec3(0.0f, 0.27f, 0.67f) * t;
		}
		return glm::vec3(0);
	}
}

int main(int argc, char* argv[]) {
	glm::vec3 LLC, horiz, vert;

	vector<ray> screenRays;
	vector<shared_ptr<Shape>> objects;
	shared_ptr<glm::mat4> objTrans;
	bool objHasTrans = false;

	ifstream inFile;
	ofstream outFile;

	camera cam;
	ray u, v, w;

	srand(time(NULL));

	if (argc < 8) {
		cerr << "Error format: a.out <width> <height> <input_filename> <output_filename> <shading_model> <numRays_perPixel> <recursionDepth> <optional_obj>" << endl;
	}
	else {

		int width = atoi(argv[1]);
		int height = atoi(argv[2]);
		int shading_mode = atoi(argv[5]);
		int rpp = atoi(argv[6]);	//rays per pixel
		int recDepth = atoi(argv[7]);	//recursion depth / ray bounces
		inFile.open(argv[3]);
		cout << "reading pov: " << argv[3] << endl;
		outFile.open(argv[4]);
		cout << "writing to: " << argv[4] << endl;
		string obje = "";
		if (argc > 8)
			obje = argv[8];

		Light sceneLight;

		bool hasEmissive = false;

		readPov(inFile, cam, sceneLight, objects, objTrans, objHasTrans);

		for (auto& s : objects) {
			cout << s->getType() << endl;
			if (s->getType() == "rectangle") {
				if (s->getMat()->getType() == "emissive")
					hasEmissive = true;
			}
			cout << endl;
		}

		glm::vec3 wDir = cam.getEye() - cam.getLA();

		//generate u,v,w
		wDir = glm::normalize(wDir);
		w = ray(cam.getEye(), wDir);
		glm::vec3 uDir = glm::cross(cam.getUp(), w.getDir());//cam.getUp().cross(w.getDir());
		u = ray(cam.getEye(), uDir);
		glm::vec3 vDir = glm::cross(w.getDir(), u.getDir());//w.getDir().cross(u.getDir());
		v = ray(cam.getEye(), vDir);

		//generate horizontal and vertical
		horiz = uDir * glm::length(cam.getRight());//cam.getRight().leng();
		vert = vDir * glm::length(cam.getUp());//cam.getUp().leng();

		glm::vec3* pixels = new glm::vec3[width * height];

		//generate LLC
		LLC = cam.getEye() - (horiz * 0.5f) - (vert * 0.5f) - w.getDir();

		glm::vec3 cummulativeCol = glm::vec3(0.0f);
		ray pixRay = ray();
		glm::vec3 unitDir = glm::vec3(0.0f);
		glm::vec3 rgb = glm::vec3(0.0f);
		float gamma = 2.0f;

		auto begin = std::chrono::high_resolution_clock::now();
		cout << "starting..." << endl;

		vector<tinyobj::shape_t> TOShapes;
		vector<tinyobj::material_t> TOMaterials;
		string errorStr;

		std::string resourceDir;

		// depth of field vars
		ray sampleRay = ray();
		glm::vec3 sRayP, sRayD, aimPoint;
		float xJit, yJit, jitAngle;
		float focalDist = 10.0f;
		float focalR = 0.1f; //normal: 0.1f
		glm::vec3 rgbs;

		if (obje != "") {
			bool loadCheck = tinyobj::LoadObj(TOShapes, TOMaterials, errorStr, obje.c_str());
			if (!loadCheck)
				throw errorStr;
			else {
				shared_ptr<Material> m = make_shared<Lambert>(glm::vec3(0.6f, 0.6f, 0.9f));
				shared_ptr<Shape> object;
				if (objHasTrans)
					object = make_shared<Obj>(TOShapes[0], m, objTrans);
				else
					object = make_shared<Obj>(TOShapes[0], m);
				objects.emplace_back(object);
			}
		}

		//loop through pixels depending on shading mode
		if (shading_mode == 0) {
			for (int y = 0; y < height; y++) {	//change to for y = h-1 -> 0 to invert
				cout << y << endl;
				for (int x = 0; x < width; x++) {
					cummulativeCol = glm::vec3(0.0f);
					//cout << "starting pixel " << x << "," << y << endl;
					for (int r = 0; r < rpp; r++) {
						float vPos = (float)y + randf() - 0.5f;
						float hPos = (float)x + randf() - 0.5f;
						pixRay = genRay(cam.getEye(), LLC, horiz, vert, width, height, hPos, vPos);

						rgb = phong(cam, sceneLight, pixRay, objects);

						cummulativeCol += rgb;
					}
					//cout << "finished pixel " << x << "," << y << endl;
					cummulativeCol /= rpp;
					cummulativeCol.x = sqrt((2.0f / gamma) * cummulativeCol.x);
					cummulativeCol.y = sqrt((2.0f / gamma) * cummulativeCol.y);
					cummulativeCol.z = sqrt((2.0f / gamma) * cummulativeCol.z);
					pixels[y * width + x] = cummulativeCol;
				}
			}
		}
		else {
			for (int y = 0; y < height; y++) {	//change to for y = h-1 -> 0 to invert
				cout << y << endl;
				for (int x = 0; x < width; x++) {
					cummulativeCol = glm::vec3(0.0f);
					//cout << "starting pixel " << x << "," << y << endl;
					for (int r = 0; r < rpp; r++) {
						float vPos = (float)y + randf() - 0.5f;
						float hPos = (float)x + randf() - 0.5f;

						pixRay = genRay(cam.getEye(), LLC, horiz, vert, width, height, hPos, vPos);

						// depth of field rendering
						aimPoint = pixRay.getPos() + focalDist * pixRay.getDir();
						for (int di = 0; di < 8; di++) {
							float du = rand() / float(RAND_MAX + 1);
							float dv = rand() / float(RAND_MAX + 1);

							jitAngle = 45.0f * di;
							xJit = cos(jitAngle) * focalR;
							yJit = sin(jitAngle) * focalR;
							sRayP = pixRay.getPos();
							sRayP.x += xJit;
							sRayP.y += yJit;

							sRayD = aimPoint - sRayP;
							glm::normalize(sRayD);

							sampleRay = ray(sRayP, sRayD);
							rgb = foggy(cam, sceneLight, sampleRay, objects, recDepth, hasEmissive, true);
							rgbs += rgb;
						}

						//rgb = foggy(cam, sceneLight, pixRay, objects, recDepth, hasEmissive, true);
						rgbs /= 8;

						cummulativeCol += rgbs;
					}
					//cout << "finished pixel " << x << "," << y << endl;
					cummulativeCol /= rpp;
					//increased from 1 to 2 for brightness adjustment
					cummulativeCol.x = sqrt((2.0f / gamma) * cummulativeCol.x);
					cummulativeCol.y = sqrt((2.0f / gamma) * cummulativeCol.y);
					cummulativeCol.z = sqrt((2.0f / gamma) * cummulativeCol.z);
					pixels[y * width + x] = cummulativeCol;
				}
			}
		}
		auto end = std::chrono::high_resolution_clock::now();
		auto elapsed = std::chrono::duration_cast<std::chrono::nanoseconds>(end - begin);
		printf("Time measured: %.3f seconds.\n", elapsed.count() * 1e-9);

		//output to .ppm file
		if (width > 0 && height > 0) {
			ppm theWriter(width, height, pixels);	//width, height, shape
			//ppm theWriter(width, height, newPixels);
			if (outFile) {
				cout << "attempting to write..." << endl;
				theWriter.writeOut(outFile);
				cout << "Done!" << endl;
			}
			else
				cerr << "error writing file" << endl;
		} //infile valid
		else {
			cout << "invalid input from file" << endl;
		}//infile invalid
	}
}
